// Source file for the off-line renderer



// include files

#include "R3Graphics/R3Graphics.h"



// Program arguments

static char *scenename = NULL;
static char *imagename = NULL;
static int imagewidth = 128;
static int imageheight = 128;
static RNBoolean print_verbose = FALSE;



static void
Render(R3Scene *scene, R2Image *image)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get camera
  const R3Camera *camera = scene->Camera();
  if (!camera) RNAbort("There is no camera in the scene");

  // Create viewport
  R2Viewport viewport(0, 0, imagewidth, imageheight);

  // Create viewer (camera-viewport mapping)
  R3Viewer viewer(*camera, viewport);

  // Generate ray through each pixel
  for (int i = 0; i < image->Width(); i++) {
    for (int j = 0; j < image->Height(); j++) {
      // Generate ray through pixel (i,j)
      R3Ray ray = viewer.WorldRay(i, j);

      // Find intersection
      R3Node *node;
      R3Point point;
      R3Vector normal;
      RNScalar t;
      if (scene->Intersects(ray, &node, &point, &normal, &t)) {
        // Set color of pixel (i,j)
        RNRgb rgb(i / 255.0, j / 255.0, 0.0);
        image->SetPixelRGB(i, j, rgb);
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf(" Elapsed time = %.6f seconds\n", start_time.Elapsed());
    printf(" Resolution = %d x %d\n", imagewidth, imageheight);
  }
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: rview scenename imagename [-res width height]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-resolution")) { 
        argv++; argc--; imagewidth = atoi(*argv); 
        argv++; argc--; imageheight = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-v")) print_verbose = TRUE;
      else RNAbort("Invalid program argument: %s", *argv);
      argv++; argc--;
    }
    else {
      if (!scenename) scenename = *argv;
      else if (!imagename) imagename = *argv;
      else RNAbort("Invalid program argument: %s", *argv);
      argv++; argc--;
    }
  }

  // Check scene filename
  if (!scenename) {
    RNFail("You did not specify a scene file name");	
    return FALSE;
  }

  // Check image filename
  if (!imagename) {
    RNFail("You did not specify a image file name");	
    return FALSE;
  }

  // Check image dimensions
  if ((imagewidth < 1) || (imageheight < 1)) {
    RNFail("You entered invalid image dimensions: %d %d", imagewidth, imageheight);
    return FALSE;
  }

  // Return OK status 
  return TRUE;
}



int 
main(int argc, char **argv)
{
  // Initialize packages
  if (!R3InitGraphics()) exit(1);

  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(1);

  // Create input scene
  R3Scene *scene = R3ReadRayFile(scenename);
  if (!scene) exit(1);

  // Create output image
  R2Image *image = new R2Image(imagewidth, imageheight);
  assert(image);

  // Render scene into image
  Render(scene, image);

  // Write image
  image->Write(imagename);

  // Stop packages
  R3StopGraphics();

  // Return success 
  return 0;
}











